import abc
from typing import List, Tuple

from .base import Indexer


class PipeIndexer(Indexer):

    def __init__(
            self,
            sos_token: str,
            eos_token: str,
            pad_token: str,
            unk_token: str,
            maxlen: int = 50,
        ):
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.pad_token = pad_token
        self.unk_token = unk_token
        self.maxlen = maxlen
        self.pipe = self._build_pipe()

    @abc.abstractmethod
    def _build_pipe(self):
        pass

    @abc.abstractproperty
    def token2index_pipe_layer(self):
        pass

    def word2index(self, word):
        token2index = self.pipe.get_state(self.token2index_pipe_layer)['token2index']
        if word not in token2index:
            raise KeyError(f'{word} is not in vocab.')
        return token2index[word]

    def index2word(self, index):
        index2token = self.pipe.get_state(self.token2index_pipe_layer)['index2token']
        if index not in index2token:
            raise KeyError(f'{index} is not in index map.')
        return index2token[index]

    def transform(
            self,
            utterances: List[str],
        ) -> Tuple[List[List[int]], dict]:
        result, tx_info, intermediates = self.pipe.transform(utterances)
        output_info = {
            'seqlen': self._compute_seqlen(intermediates[0], maxlen=self.maxlen),
            'inv_info': tx_info,
        }
        return result, output_info

    @staticmethod
    def _compute_seqlen(
            sentences: List[List[str]],
            maxlen: int,
        ) -> List[int]:
        output = [0] * len(sentences)
        for i, sent in enumerate(sentences):
            output[i] = min(len(sent), maxlen)
        return output

    def inverse_transform(
            self,
            indices: List[List[int]],
            tx_info: List[dict],
        ) -> List[str]:
        return self.pipe.inverse_transform(indices, tx_info)



import abc
import os
from pathlib import Path
import shutil


def list_all_files(root: str):
    output = []
    for prefix, _, files in os.walk(root):
        for f in files:
            path = os.path.join(prefix, f)
            output.append(path)
    return output


class TestTemplate(abc.ABC):

    @classmethod
    def setUpClass(cls):
        cls.sos_token = '<sos>'
        cls.eos_token = '</s>'
        cls.pad_token = '<pad>'
        cls.unk_token = '<unk>'
        cls.maxlen = 7
        cls.input_data = [
            '克安是牛肉大粉絲',  # longer than 7 after adding sos eos
            '繼良喜歡喝星巴巴',  # longer than 7 after adding sos eos
            '安靜的祥睿',  # equal to 7 after adding sos eos
            '喔',  # shorter than 7 after adding sos eos
        ]
        cls.output_dir = Path(__file__).parent / 'example_indexer/'

    def setUp(self):
        self.indexer = self.get_indexer()
        self.indexer.fit(self.input_data)

    def tearDown(self):
        if self.output_dir.exists():
            shutil.rmtree(str(self.output_dir))

    @abc.abstractmethod
    def get_indexer_class(self):
        pass

    @abc.abstractmethod
    def get_indexer(self):
        pass

    @abc.abstractmethod
    def get_correct_idxs_and_seqlen_of_input_data(self):
        pass

    def test_word2index_out_of_range(self):
        with self.assertRaises(KeyError):
            self.indexer.word2index('凢')

    def test_index2word_out_of_range(self):
        with self.assertRaises(KeyError):
            self.indexer.index2word(100000000000)

    def test_transform(self):
        tx_data, meta = self.indexer.transform(self.input_data)
        correct_idxs, correct_seqs = self.get_correct_idxs_and_seqlen_of_input_data()
        self.assertEqual(correct_idxs, tx_data)
        self.assertEqual(correct_seqs, meta['seqlen'])

    def test_inverse_transform(self):
        tx_data, meta = self.indexer.transform(self.input_data)
        output = self.indexer.inverse_transform(tx_data, meta['inv_info'])
        self.assertEqual(output, self.input_data)

    def test_save(self):
        output_dir = str(self.output_dir)
        self.indexer.save(output_dir)
        self.assertEqual(
            set([os.path.join(output_dir, filepath) for filepath in ['pipe.json', 'indexer.json']]),
            set(list_all_files(output_dir)),
        )

    def test_load(self):
        self.indexer.save(str(self.output_dir))
        indexer = self.get_indexer_class().load(str(self.output_dir))
        tx_data, meta = indexer.transform(self.input_data)
        correct_idxs, correct_seqs = self.get_correct_idxs_and_seqlen_of_input_data()
        self.assertEqual(correct_idxs, tx_data)
        self.assertEqual(correct_seqs, meta['seqlen'])



import os
import errno
import json


def save_json(data, path):
    with open(path, 'w', encoding='utf-8') as filep:
        json.dump(data, filep, ensure_ascii=False, indent=2)


def load_json(path):
    with open(path, 'r', encoding='utf-8') as filep:
        output = json.load(filep)
    return output


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



from typing import List, Tuple

from abc import abstractmethod, abstractclassmethod, ABC


class Indexer(ABC):

    @abstractmethod
    def fit(self, utterances: List[str]):
        pass

    @abstractmethod
    def transform(
            self,
            data: List[str],
        ) -> Tuple[List[List[int]], dict]:
        """Transform strings to indices"""
        pass

    @abstractmethod
    def inverse_transform(
            self,
            data: List[List[int]],
        ) -> List[str]:
        """Restore indices to strings"""
        pass

    @abstractmethod
    def save(self, output_path: str):
        pass

    @abstractclassmethod
    def load(cls, output_path: str):
        pass



from unittest import TestCase

from pathlib import Path
import umsgpack

from .template import TestTemplate
from ..char_indexer import CharIndexer


def load_w2v(path: str):
    with open(path, 'rb') as filep:
        word2vec = umsgpack.unpack(filep)
    return word2vec


class CharIndexerWithoutW2vTestCase(TestTemplate, TestCase):

    def get_indexer_class(self):
        return CharIndexer

    def get_indexer(self):
        return CharIndexer.create_without_word2vec(
            sos_token=self.sos_token,
            eos_token=self.eos_token,
            pad_token=self.pad_token,
            unk_token=self.unk_token,
            maxlen=self.maxlen,
        )

    def get_correct_idxs_and_seqlen_of_input_data(self):
        correct_idxs = []
        for sent in self.input_data:
            sent_idxs = [self.indexer.word2index(self.indexer.sos_token)]
            for word in sent:
                try:
                    sent_idxs.append(
                        self.indexer.word2index(word),
                    )
                except KeyError:
                    sent_idxs.append(
                        self.indexer.word2index(self.indexer.unk_token),
                    )
            sent_idxs.append(self.indexer.word2index(self.indexer.eos_token))
            if len(sent_idxs) > self.maxlen:
                sent_idxs = sent_idxs[:self.maxlen]
            while len(sent_idxs) < self.maxlen:
                sent_idxs.append(self.indexer.word2index(self.indexer.pad_token))
            assert len(sent_idxs) == self.maxlen
            correct_idxs.append(sent_idxs)

        correct_seqs = [
            min(len(sent) + 2, self.maxlen)
            for sent in self.input_data
        ]
        return correct_idxs, correct_seqs

    def test_embedding_correct(self):
        self.assertIsNone(self.indexer.word2vec)


class CharIndexerWithW2vTestCase(CharIndexerWithoutW2vTestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_emb = load_w2v(
            Path(__file__).resolve().parent.joinpath('data/example.msg'),
        )
        super().setUpClass()

    def get_indexer(self):
        return CharIndexer.create_with_word2vec(
            word2vec=self.test_emb,
            sos_token=self.sos_token,
            eos_token=self.eos_token,
            pad_token=self.pad_token,
            unk_token=self.unk_token,
            maxlen=self.maxlen,
        )

    def test_embedding_correct(self):
        self.assertEqual(self.indexer.word2vec, self.test_emb)

    def test_transform_and_fit_dont_change(self):
        tx_data, meta = self.indexer.transform(self.input_data)
        correct_idxs, correct_seqs = self.get_correct_idxs_and_seqlen_of_input_data()
        self.assertEqual(correct_idxs, tx_data)
        self.assertEqual(correct_seqs, meta['seqlen'])
        self.indexer.fit(self.input_data)
        self.assertEqual(correct_idxs, tx_data)
        self.assertEqual(correct_seqs, meta['seqlen'])



from typing import List
from unittest import TestCase
from unittest.mock import patch
import shutil
from os.path import join, abspath, exists, dirname
import os

from ..io import save_indexer, load_indexer
from text_indexer.indexers.utils import save_json, load_json


def list_all_files(root: str) -> List[str]:
    output = []
    for prefix, _, files in os.walk(root):
        for f in files:
            path = os.path.join(prefix, f)
            output.append(path)
    return output


class MockIndexer(object):

    pipe_filename = 'fake_pipe.json'
    indexer_filename = 'fake_indexer.json'

    def __init__(self, aa=1, bb=2):
        self.aa = aa
        self.bb = bb
        self.a = 1
        self.b = 2

    def save(self, output_dir):
        save_json({'a': self.a, 'b': self.b}, join(output_dir, self.pipe_filename))
        save_json({'aa': self.aa, 'bb': self.bb}, join(output_dir, self.indexer_filename))

    @classmethod
    def load(cls, output_dir):
        pipe = load_json(join(output_dir, cls.pipe_filename))
        params = load_json(join(output_dir, cls.indexer_filename))
        indexer = cls(**params)
        indexer.pipe = pipe
        return indexer

    def transform(self, utterances):
        return [f"{self.aa}_{self.bb}_{utt}" for utt in utterances]


class IOTestCase(TestCase):

    def setUp(self):
        root_dir = dirname(abspath(__file__))
        self.output_dir = join(root_dir, 'example/')
        os.mkdir(self.output_dir)

    def tearDown(self):
        if exists(self.output_dir):
            shutil.rmtree(self.output_dir)

    def test_save_indexer(self):
        save_indexer(indexer=MockIndexer(), output_dir=self.output_dir)
        self.assertTrue(exists(self.output_dir))
        self.assertEqual(
            set(
                [
                    os.path.join(self.output_dir, MockIndexer.pipe_filename),
                    os.path.join(self.output_dir, MockIndexer.indexer_filename),
                    os.path.join(self.output_dir, 'name'),
                ],
            ),
            set(list_all_files(self.output_dir)),
        )

    def test_load_indexer(self):
        save_indexer(indexer=MockIndexer(), output_dir=self.output_dir)
        with patch('text_indexer.io._get_indexer_module', return_value=MockIndexer):
            indexer = load_indexer(self.output_dir)
        output = indexer.transform(['dummy_utt'])
        self.assertEqual([f'{indexer.aa}_{indexer.bb}_dummy_utt'], output)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    long_description = open("README.md").read()
except IOError:
    long_description = ""


setup(
    name='text-indexer',
    version='1.1.1',
    description="Yoctol Text Indexer",
    url="https://github.com/Yoctol/text-indexer",
    license="MIT",
    author="Yoctol Info",
    author_email='yien.tsai@yoctol.com',
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        "strpipe==0.4.7",
        "u-msgpack-python==2.5.0",
    ],
    include_package_data=True,
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
    ],
)



import os
from os.path import join, isdir
import logging

from .indexers import (
    Indexer,
    CharIndexer,
)


LOGGER = logging.getLogger('__file__')
INDEXERS = {
    indexer_cls.__class__.__name__: indexer_cls for indexer_cls in [
        CharIndexer,
    ]
}


def save_indexer(
        indexer: Indexer,
        output_dir: str,
        logger: logging.Logger = LOGGER,
    ) -> str:

    if not isdir(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    # save indexer class name
    class_name = indexer.__class__.__name__
    logger.info(f'Saving indexer {class_name} to {output_dir}')
    _save_name(class_name, _gen_name_path(output_dir))

    # save indexer
    indexer.save(output_dir)  # save indexer
    del indexer


def load_indexer(
        output_dir: str,
        logger: logging.Logger = LOGGER,
    ) -> Indexer:

    _validate_dir(output_dir)
    logger.info(f'Loading indexer from {output_dir}')

    # load indexer
    indexer_name = _load_name(_gen_name_path(output_dir))
    indexer_module = _get_indexer_module(indexer_name)
    indexer = indexer_module.load(output_dir)

    return indexer


def _validate_dir(directory: str):
    if not isdir(directory):
        raise ValueError(f'{directory} is not a directory.')


def _save_name(name: str, path: str) -> None:
    with open(path, 'w', encoding='utf-8') as text_file:
        text_file.write(name)


def _load_name(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as text_file:
        name = text_file.read()
    return name


def _gen_name_path(directory: str) -> str:
    return join(directory, 'name')


def _get_indexer_module(indexer_name: str) -> Indexer:
    return INDEXERS[indexer_name]



from typing import List
import os
import warnings

import strpipe as sp

from .base import Indexer
from .pipe_indexer import PipeIndexer
from .utils import load_json, save_json, mkdir_p


class CharIndexer(PipeIndexer):

    def __init__(
            self,
            word2vec=None,
            sos_token: str = '<sos>',
            eos_token: str = '</s>',
            pad_token: str = '<pad>',
            unk_token: str = '<unk>',
            maxlen: int = 50,
        ):
        self.word2vec = word2vec
        self._token2index_pipe_layer = None
        super().__init__(
            sos_token=sos_token,
            eos_token=eos_token,
            pad_token=pad_token,
            unk_token=unk_token,
            maxlen=maxlen,
        )

    @classmethod
    def create_without_word2vec(
            cls,
            sos_token: str = '<sos>',
            eos_token: str = '</s>',
            pad_token: str = '<pad>',
            unk_token: str = '<unk>',
            maxlen: int = 50,
        ):
        indexer = cls(
            sos_token=sos_token,
            eos_token=eos_token,
            pad_token=pad_token,
            unk_token=unk_token,
            maxlen=maxlen,
        )
        return indexer

    @classmethod
    def create_with_word2vec(
            cls,
            word2vec,
            sos_token: str = '<sos>',
            eos_token: str = '</s>',
            pad_token: str = '<pad>',
            unk_token: str = '<unk>',
            maxlen: int = 50,
        ):
        indexer = cls(
            word2vec=word2vec,
            sos_token=sos_token,
            eos_token=eos_token,
            pad_token=pad_token,
            unk_token=unk_token,
            maxlen=maxlen,
        )
        return indexer

    def _build_pipe(self):
        p = sp.Pipe()
        p.add_step_by_op_name('CharTokenizer')
        p.add_step_by_op_name(
            'AddSosEos',
            op_kwargs={
                'sos_token': self.sos_token,
                'eos_token': self.eos_token,
            },
        )
        p.add_checkpoint()
        p.add_step_by_op_name(
            'Pad',
            op_kwargs={
                'pad_token': self.pad_token,
                'maxlen': self.maxlen,
            },
        )
        if self.word2vec is None:
            p.add_step_by_op_name(
                'TokenToIndex',
                op_kwargs={
                    'unk_token': self.unk_token,
                },
            )
        else:
            p.add_step_by_op_name(
                'TokenToIndex',
                op_kwargs={
                    'unk_token': self.unk_token,
                    'token2index': self.word2vec['token2index'],
                },
            )
        self._token2index_pipe_layer = 3
        return p

    @property
    def token2index_pipe_layer(self):
        return self._token2index_pipe_layer

    def fit(self, utterances: List[str]):
        if self.word2vec is None:
            self.pipe.fit(utterances)
        else:
            warnings.warn(
                "CharwtWord2Vec fit function doesn't actually fit on utterances.",
                UserWarning,
            )
            self.pipe.fit(['dummy fit'])

    def save(self, output_dir: str):
        mkdir_p(output_dir)
        params = {
            "maxlen": self.maxlen,
            "sos_token": self.sos_token,
            "eos_token": self.eos_token,
            "pad_token": self.pad_token,
            "unk_token": self.unk_token,
        }
        save_json(params, os.path.join(output_dir, 'indexer.json'))
        self.pipe.save_json(os.path.join(output_dir, 'pipe.json'))

    @classmethod
    def load(cls, output_dir: str) -> Indexer:
        params = load_json(os.path.join(output_dir, 'indexer.json'))
        indexer = cls.create_without_word2vec(**params)
        indexer.pipe = sp.Pipe.restore_from_json(os.path.join(output_dir, 'pipe.json'))
        return indexer
